@@ -1,8 +1,10 @@
 #![allow(unused_variables)]
 
+use log;
+
 use crate::language::FloatType;
-use crate::parser::stmt::{StmtMeta, Stmt, Label};
-use crate::parser::expr::{Expr, ExprMeta};
+use crate::parser::stmt::{StmtMeta, Stmt, Label, StmtList, ControlFlow};
+use crate::parser::expr::{Expr, ExprMeta, Conditional};
 use crate::parser::primary::{Atom, Primary};
 use crate::parser::assign::{Assignment, Declaration, LValue, DeclType};
 use crate::runtime::types::operator::{UnaryOp, BinaryOp, Arithmetic, Bitwise, Shift, Comparison, Logical};
@@ -81,7 +83,7 @@ struct Local {
     offset: Offset,
 }
 
-#[derive(Debug, Clone, PartialEq, Eq)]
+#[derive(Debug, Clone, Copy, PartialEq, Eq)]
 enum ScopeTag {
     Block,
     Loop,
@@ -90,6 +92,7 @@ enum ScopeTag {
     Class,
 }
 
+
 #[derive(Debug)]
 struct Scope {
     tag: ScopeTag,
@@ -125,6 +128,15 @@ impl Scope {
         self.locals.push(local);
         Ok(self.locals.last().unwrap())
     }
+    
+    // i.e., without the "nonlocal" keyword
+    // this returns true if we can assign into the enclosing scope
+    fn allow_enclosing_assignment(&self) -> bool {
+        match self.tag {
+            ScopeTag::Loop | ScopeTag::Branch => true,
+            _ => false,
+        }
+    }
 }
 
 #[derive(Debug)]
@@ -177,9 +189,44 @@ impl CompilerState {
         Ok(())
     }
     
-    fn resolve_local(&mut self, name: &InternSymbol) -> Option<&Local> {
+    // find the nearest local in any scope
+    fn resolve_local(&self, name: &InternSymbol) -> Option<&Local> {
         self.scopes.iter().rev().find_map(|scope| scope.find_local(name))
     }
+    
+    // find the nearest local in scopes that allow nonlocal assignment
+    fn resolve_local_strict(&self, name: &InternSymbol) -> Option<&Local> {
+        for scope in self.scopes.iter().rev() {
+            let local = scope.find_local(name);
+            if local.is_some() {
+                return local;
+            }
+            
+            if !scope.allow_enclosing_assignment() {
+                break;
+            }
+        }
+        
+        None
+    }
+    
+    // search scopes that would have been skipped by resolve_local_strict()
+    fn resolve_nonlocal(&self, name: &InternSymbol) -> Option<&Local> {
+        let mut is_local = true;
+        for scope in self.scopes.iter().rev() {
+            if is_local {
+                is_local &= scope.allow_enclosing_assignment();
+                continue;
+            }
+            
+            let local = scope.find_local(name);
+            if local.is_some() {
+                return local;
+            }
+        }
+        
+        None
+    }
 }
 
 
@@ -230,27 +277,44 @@ impl CodeGenerator {
         }
     }
     
-    fn emit_instr(&mut self, symbol: &DebugSymbol, opcode: OpCode) -> CompileResult<()> {
+    fn emit_instr(&mut self, symbol: &DebugSymbol, opcode: OpCode) {
         debug_assert!(opcode.instr_len() == 1);
         self.symbols.push(symbol);
         self.chunk.push_byte(opcode);
-        Ok(())
     }
     
-    fn emit_instr_byte(&mut self, symbol: &DebugSymbol, opcode: OpCode, byte: u8) -> CompileResult<()> {
+    fn emit_instr_byte(&mut self, symbol: &DebugSymbol, opcode: OpCode, byte: u8) {
         debug_assert!(opcode.instr_len() == 2);
         self.symbols.push(symbol);
         self.chunk.push_byte(opcode);
         self.chunk.push_byte(byte);
-        Ok(())
     }
     
-    fn emit_instr_data<const N: usize>(&mut self, symbol: &DebugSymbol, opcode: OpCode, bytes: [u8; N]) -> CompileResult<()> {
-        debug_assert!(opcode.instr_len() == 1 + N);
+    fn emit_instr_data(&mut self, symbol: &DebugSymbol, opcode: OpCode, bytes: &[u8]) {
+        debug_assert!(opcode.instr_len() == 1 + bytes.len());
         self.symbols.push(symbol);
         self.chunk.push_byte(opcode);
         self.chunk.extend_bytes(&bytes);
-        Ok(())
+    }
+    
+    // returns the offset to end of the instruction, for use with jump site patching
+    fn emit_dummy_instr(&mut self, symbol: &DebugSymbol, opcode: OpCode) -> usize {
+        self.symbols.push(symbol);
+        
+        let bytes = std::iter::repeat(OpCode::Nop).take(opcode.instr_len());
+        for byte in bytes {
+            self.chunk.push_byte(byte);
+        }
+        self.chunk.bytes().len()
+    }
+    
+    // the given offset is the offset that is expected to be the *end* of the patched instruction
+    fn patch_instr_data<const N: usize>(&mut self, end_offset: usize, opcode: OpCode, bytes: &[u8; N]) {
+        debug_assert!(opcode.instr_len() == 1 + N);
+        
+        let offset = end_offset - opcode.instr_len();
+        self.chunk.bytes_mut()[offset] = u8::from(opcode);
+        self.chunk.patch_bytes(offset + 1, bytes);
     }
     
     fn make_const(&mut self, symbol: &DebugSymbol, value: Constant) -> CompileResult<ConstID> {
@@ -258,100 +322,176 @@ impl CodeGenerator {
             .map_err(|error| error.with_symbol(*symbol))
     }
     
-    fn emit_const(&mut self, symbol: &DebugSymbol, value: Constant) -> CompileResult<()> {
+    fn emit_load_const(&mut self, symbol: &DebugSymbol, value: Constant) -> CompileResult<()> {
         let cid = self.make_const(symbol, value)?;
         
         if cid <= u8::MAX.into() {
-            self.emit_instr_byte(symbol, OpCode::LoadConst, u8::try_from(cid).unwrap())
+            self.emit_instr_byte(symbol, OpCode::LoadConst, u8::try_from(cid).unwrap());
         } else {
-            self.emit_instr_data(symbol, OpCode::LoadConst16, cid.to_le_bytes())
+            self.emit_instr_data(symbol, OpCode::LoadConst16, &cid.to_le_bytes());
         }
-    }
-    
-    fn emit_begin_scope(&mut self, tag: ScopeTag, symbol: &DebugSymbol) -> CompileResult<()> {
-        self.state.push_scope(tag, *symbol);
         Ok(())
     }
     
-    fn emit_end_scope(&mut self) -> CompileResult<()> {
+    fn emit_begin_scope(&mut self, tag: ScopeTag, symbol: &DebugSymbol) {
+        self.state.push_scope(tag, *symbol);
+    }
+    
+    fn emit_end_scope(&mut self) {
         let scope = self.state.pop_scope();
         let symbol = scope.symbol;
         
         // discard all the locals from the stack
         let mut discard = scope.locals.len();
         while discard > u8::MAX.into() {
-            self.emit_instr_byte(&symbol, OpCode::DropLocals, u8::MAX)?;
+            self.emit_instr_byte(&symbol, OpCode::DropLocals, u8::MAX);
             discard -= usize::from(u8::MAX);
         }
         
-        self.emit_instr_byte(&symbol, OpCode::DropLocals, u8::try_from(discard).unwrap())?;
-        Ok(())
+        if discard > 0 {
+            self.emit_instr_byte(&symbol, OpCode::DropLocals, u8::try_from(discard).unwrap());
+        }
     }
     
     fn compile_stmt(&mut self, symbol: &DebugSymbol, stmt: &Stmt) -> CompileResult<()> {
         match stmt {
             Stmt::Echo(expr) => {
                 self.compile_expr(symbol, expr)?;
-                self.emit_instr(symbol, OpCode::Inspect)
+                self.emit_instr(symbol, OpCode::Inspect);
             },
             Stmt::Assert(expr) => {
                 self.compile_expr(symbol, expr)?;
-                self.emit_instr(symbol, OpCode::Assert)
+                self.emit_instr(symbol, OpCode::Assert);
             }
             
             Stmt::Expression(expr) => {
                 self.compile_expr(symbol, expr)?;
-                self.emit_instr(symbol, OpCode::Pop)
+                self.emit_instr(symbol, OpCode::Pop);
             },
-            
-            Stmt::Continue(label) => unimplemented!(),
-            Stmt::Break(label, expr) => unimplemented!(),
-            Stmt::Return(expr) => unimplemented!(),
         }
+        Ok(())
     }
     
     fn compile_expr(&mut self, symbol: &DebugSymbol, expr: &Expr) -> CompileResult<()> {
         match expr {
-            Expr::Atom(atom) => self.compile_atom(symbol, atom),
+            Expr::Atom(atom) => self.compile_atom(symbol, atom)?,
             
             Expr::Primary(primary) => unimplemented!(),
             
             Expr::UnaryOp(op, expr) => {
                 self.compile_expr(symbol, expr)?;
-                self.emit_unary_op(symbol, op)
+                self.emit_unary_op(symbol, op);
             },
             
             Expr::BinaryOp(op, exprs) => {
                 let (ref lhs, ref rhs) = **exprs;
                 self.compile_expr(symbol, lhs)?;
                 self.compile_expr(symbol, rhs)?;
-                self.emit_binary_op(symbol, op)
+                self.emit_binary_op(symbol, op);
             },
             
-            Expr::Declaration(decl) => self.compile_declaration(symbol, (&**decl).into()),
-            Expr::Assignment(assign) => self.compile_assignment(symbol, (&**assign).into()),
+            Expr::Declaration(decl) => self.compile_declaration(symbol, (&**decl).into())?,
+            Expr::Assignment(assign) => self.compile_assignment(symbol, (&**assign).into())?,
             
-            Expr::Tuple(expr_list) => self.compile_tuple(symbol, expr_list),
+            Expr::Tuple(expr_list) => self.compile_tuple(symbol, expr_list)?,
             
-            Expr::Block(label, suite) => self.compile_block(symbol, label.as_ref(), suite),
-            Expr::IfExpr(if_expr) => unimplemented!(),
+            Expr::Block(label, stmt_list) => self.compile_block(symbol, label.as_ref(), stmt_list)?,
+            Expr::IfExpr(cond) => self.compile_if_expr(symbol, cond)?,
             
             Expr::FunctionDef(fundef) => unimplemented!(),
         }
+        Ok(())
     }
     
-    fn compile_block(&mut self, symbol: &DebugSymbol, label: Option<&Label>, suite: &[StmtMeta]) -> CompileResult<()> {
-        self.emit_begin_scope(ScopeTag::Block, symbol)?;
+    fn compile_block(&mut self, symbol: &DebugSymbol, label: Option<&Label>, stmt_list: &StmtList) -> CompileResult<()> {
         
-        // TODO control flow
-        for stmt in suite.iter() {
+        self.compile_stmt_list(ScopeTag::Block, symbol, stmt_list)?;
+        
+        Ok(())
+    }
+    
+    fn compile_if_expr(&mut self, symbol: &DebugSymbol, conditional: &Conditional) -> CompileResult<()> {
+        debug_assert!(!conditional.branches().is_empty());
+        
+        // track the sites where we jump to the end, so we can patch them later
+        let mut end_jump_sites = Vec::new();
+        
+        for (index, branch) in conditional.branches().iter().enumerate() {
+            
+            self.compile_expr(symbol, branch.cond_expr())?;
+            let branch_jump_site = self.emit_dummy_instr(symbol, OpCode::PopJumpIfFalse);
+            
+            self.compile_stmt_list(ScopeTag::Branch, symbol, branch.suite())?;
+            
+            // site for the jump to end
+            let end_offset = self.emit_dummy_instr(symbol, OpCode::Jump);
+            end_jump_sites.push(end_offset);
+            
+            // target for the jump from the conditional of the now compiled branch
+            let branch_target = self.chunk.bytes().len();
+            let jump_offset = i16::try_from(branch_target - branch_jump_site).expect("exceeded max jump offset");
+            self.patch_instr_data(branch_jump_site, OpCode::PopJumpIfFalse, &jump_offset.to_le_bytes());
+        }
+        
+        if let Some(suite) = &conditional.else_branch() {
+            
+            self.compile_stmt_list(ScopeTag::Branch, symbol, suite)?;
+            
+        } else {
+            
+            // since if-expressions are expressions, there is always an implicit else
+            self.emit_instr(symbol, OpCode::Nil);
+            
+        }
+        
+        // patch all of the end jump sites
+        let end_target = self.chunk.bytes().len();
+        for jump_site in end_jump_sites.iter() {
+            
+            let jump_offset = i16::try_from(end_target - jump_site).expect("exceeded max jump offset");
+            self.patch_instr_data(*jump_site, OpCode::Jump, &jump_offset.to_le_bytes());
+        }
+        
+        Ok(())
+    }
+    
+    fn compile_stmt_list(&mut self, tag: ScopeTag, symbol: &DebugSymbol, stmt_list: &StmtList) -> CompileResult<()> {
+        
+        self.emit_begin_scope(tag, symbol);
+        
+        // compile stmt suite
+        for stmt in stmt_list.suite().iter() {
             let inner_symbol = stmt.debug_symbol();
             self.compile_stmt(inner_symbol, stmt.variant())?;
         }
         
-        self.emit_end_scope()?;
+        // handle control flow
+        let is_expr_block = matches!(tag, ScopeTag::Block | ScopeTag::Branch);
+        match stmt_list.control() {
+            None => if is_expr_block {
+                self.emit_instr(symbol, OpCode::Nil); // implicit nil
+            },
+            
+            Some(control) => match control {
+                ControlFlow::Expression(expr) => if is_expr_block {
+                    self.compile_expr(expr.debug_symbol(), expr.variant())?;
+                }
+                
+                ControlFlow::Continue(label) => {
+                    unimplemented!()
+                }
+                ControlFlow::Break(label, expr) => {
+                    unimplemented!()
+                }
+                ControlFlow::Return(expr) => {
+                    unimplemented!()
+                }
+            }
+        }
         
-        self.emit_instr(symbol, OpCode::Nil) // implicit nil if we don't break out of block
+        self.emit_end_scope();
+        
+        Ok(())
     }
     
     fn compile_declaration(&mut self, symbol: &DebugSymbol, decl: DeclarationRef) -> CompileResult<()> {
@@ -385,6 +525,11 @@ impl CodeGenerator {
                         self.compile_declaration(inner_symbol, inner_decl)?;
                     }
                     
+                    // declarations are also expressions
+                    let tuple_len = u8::try_from(init_list.len())
+                        .map_err(|_| CompileError::from(ErrorKind::TupleLengthLimit).with_symbol(*symbol))?;
+                    self.emit_instr_byte(symbol, OpCode::Tuple, tuple_len);
+                    
                     Ok(())
                 },
                 
@@ -398,18 +543,20 @@ impl CodeGenerator {
         self.compile_expr(symbol, init)?;  // make sure to evaluate initializer first in order for shadowing to work
         
         self.state.insert_local(decl, name).map_err(|err| err.with_symbol(*symbol))?;
-        self.emit_instr(symbol, OpCode::InsertLocal)
+        self.emit_instr(symbol, OpCode::InsertLocal);
+        Ok(())
     }
     
     fn compile_decl_global_name(&mut self, symbol: &DebugSymbol, decl: DeclType, name: InternSymbol, init: &Expr) -> CompileResult<()> {
         
         self.compile_expr(symbol, init)?;
         
-        self.emit_const(symbol, Constant::from(name))?;
+        self.emit_load_const(symbol, Constant::from(name))?;
         match decl {
             DeclType::Immutable => self.emit_instr(symbol, OpCode::InsertGlobal),
             DeclType::Mutable => self.emit_instr(symbol, OpCode::InsertGlobalMut),
         }
+        Ok(())
     }
     
     fn compile_assignment(&mut self, symbol: &DebugSymbol, assign: AssignmentRef) -> CompileResult<()> {
@@ -445,6 +592,11 @@ impl CodeGenerator {
                         self.compile_assignment(inner_symbol, inner_assign)?;
                     }
                     
+                    // assignments are also expressions
+                    let tuple_len = u8::try_from(rhs_list.len())
+                        .map_err(|_| CompileError::from(ErrorKind::TupleLengthLimit).with_symbol(*symbol))?;
+                    self.emit_instr_byte(symbol, OpCode::Tuple, tuple_len);
+                    
                     Ok(())
                 },
                 
@@ -462,7 +614,7 @@ impl CodeGenerator {
             // update assignment
             self.compile_name_lookup(symbol, name)?;
             self.compile_expr(symbol, &assign.rhs)?;
-            self.emit_binary_op(symbol, &op)?;
+            self.emit_binary_op(symbol, &op);
             
         } else {
             // normal assignment
@@ -476,7 +628,7 @@ impl CodeGenerator {
             
             // check if the name is found in the local scope...
             let local_offset = 
-                if let Some(local) = self.state.resolve_local(name) {
+                if let Some(local) = self.state.resolve_local_strict(name) {
                     if local.decl != DeclType::Mutable {
                         return Err(CompileError::from(ErrorKind::CantAssignImmutable).with_symbol(*symbol));
                     }
@@ -485,9 +637,9 @@ impl CodeGenerator {
             
             if let Some(offset) = local_offset {
                 if let Ok(offset) = u8::try_from(offset) {
-                    self.emit_instr_byte(symbol, OpCode::StoreLocal, offset)?;
+                    self.emit_instr_byte(symbol, OpCode::StoreLocal, offset);
                 } else {
-                    self.emit_instr_data(symbol, OpCode::StoreLocal16, offset.to_le_bytes())?;
+                    self.emit_instr_data(symbol, OpCode::StoreLocal16, &offset.to_le_bytes());
                 }
                 return Ok(());
             }
@@ -504,8 +656,9 @@ impl CodeGenerator {
         }
         
         // ...finally, try to assign global
-        self.emit_const(symbol, Constant::from(*name))?;
-        self.emit_instr(symbol, OpCode::StoreGlobal)
+        self.emit_load_const(symbol, Constant::from(*name))?;
+        self.emit_instr(symbol, OpCode::StoreGlobal);
+        Ok(())
     }
     
     
@@ -518,7 +671,8 @@ impl CodeGenerator {
             self.compile_expr(inner_symbol, expr.variant())?;
         }
         
-        self.emit_instr_byte(symbol, OpCode::Tuple, len)
+        self.emit_instr_byte(symbol, OpCode::Tuple, len);
+        Ok(())
     }
     
     fn compile_atom(&mut self, symbol: &DebugSymbol, atom: &Atom) -> CompileResult<()> {
@@ -530,31 +684,32 @@ impl CodeGenerator {
             
             Atom::IntegerLiteral(value) => {
                 if let Ok(value) = u8::try_from(*value) {
-                    self.emit_instr_byte(symbol, OpCode::UInt8, value)
+                    self.emit_instr_byte(symbol, OpCode::UInt8, value);
                 } else if let Ok(value) = i8::try_from(*value) {
-                    self.emit_instr_byte(symbol, OpCode::Int8, value.to_le_bytes()[0])
+                    self.emit_instr_byte(symbol, OpCode::Int8, value.to_le_bytes()[0]);
                 } else {
-                    self.emit_const(symbol, Constant::from(*value))
+                    self.emit_load_const(symbol, Constant::from(*value))?;
                 }
             },
             
             Atom::FloatLiteral(value) => {
                 if FloatType::from(i8::MIN) <= *value && *value <= FloatType::from(i8::MAX) {
                     let value = *value as i8;
-                    self.emit_instr_byte(symbol, OpCode::Float8, value.to_le_bytes()[0])
+                    self.emit_instr_byte(symbol, OpCode::Float8, value.to_le_bytes()[0]);
                 } else {
-                    self.emit_const(symbol, Constant::from(*value))
+                    self.emit_load_const(symbol, Constant::from(*value))?;
                 }
             },
-            Atom::StringLiteral(value) => self.emit_const(symbol, Constant::from(*value)),
             
-            Atom::Identifier(name) => self.compile_name_lookup(symbol, name),
+            Atom::StringLiteral(value) => self.emit_load_const(symbol, Constant::from(*value))?,
+            Atom::Identifier(name) => self.compile_name_lookup(symbol, name)?,
             
             // Atom::Self_ => unimplemented!(),
             // Atom::Super => unimplemented!(),
             
-            Atom::Group(expr) => self.compile_expr(symbol, expr),
+            Atom::Group(expr) => self.compile_expr(symbol, expr)?,
         }
+        Ok(())
     }
     
     fn compile_name_lookup(&mut self, symbol: &DebugSymbol, name: &InternSymbol) -> CompileResult<()> {
@@ -564,24 +719,25 @@ impl CodeGenerator {
             
             // Local Variable
             if let Ok(offset) = u8::try_from(offset) {
-                self.emit_instr_byte(symbol, OpCode::LoadLocal, offset)
+                self.emit_instr_byte(symbol, OpCode::LoadLocal, offset);
             } else {
-                self.emit_instr_data(symbol, OpCode::LoadLocal16, offset.to_le_bytes())
+                self.emit_instr_data(symbol, OpCode::LoadLocal16, &offset.to_le_bytes());
             }
         
         } else {
             
             // Global variable
-            self.emit_const(symbol, Constant::from(*name))?;
-            self.emit_instr(symbol, OpCode::LoadGlobal)
+            self.emit_load_const(symbol, Constant::from(*name))?;
+            self.emit_instr(symbol, OpCode::LoadGlobal);
         }
+        Ok(())
     }
     
     fn compile_primary(&mut self, symbol: &DebugSymbol, primary: &Primary) -> CompileResult<()> {
         unimplemented!()
     }
     
-    fn emit_unary_op(&mut self, symbol: &DebugSymbol, op: &UnaryOp) -> CompileResult<()> {
+    fn emit_unary_op(&mut self, symbol: &DebugSymbol, op: &UnaryOp) {
         match op {
             UnaryOp::Neg => self.emit_instr(symbol, OpCode::Neg),
             UnaryOp::Pos => self.emit_instr(symbol, OpCode::Pos),
@@ -590,7 +746,7 @@ impl CodeGenerator {
         }
     }
     
-    fn emit_binary_op(&mut self, symbol: &DebugSymbol, op: &BinaryOp) -> CompileResult<()> {
+    fn emit_binary_op(&mut self, symbol: &DebugSymbol, op: &BinaryOp) {
         match op {
             BinaryOp::Logical(logic) => unimplemented!(),
             
