@@ -133,6 +133,16 @@ impl<'c, 's> Disassembler<'c, 's> {
                     self.write_const(&mut line, &value)?;
                 }
                 
+                OpCode::Jump           |
+                OpCode::JumpIfFalse    |
+                OpCode::JumpIfTrue     |
+                OpCode::PopJumpIfFalse |
+                OpCode::PopJumpIfTrue  => {
+                    let jmp = i16::from_le_bytes(instr[1..=2].try_into().unwrap());
+                    let dest = i64::from(jmp) + i64::try_from(offset + opcode.instr_len()).expect("offset too large");
+                    write!(line, "{:16} {: >4} -> {:04X}", opcode, jmp, dest)?;
+                }
+                
                 opcode => write!(line, "{:16}", opcode)?,
             },
             
