@@ -16,7 +16,7 @@ use crate::source::{ModuleSource, SourceText};
 // will probably be attached to the statement level
 
 
-#[derive(Debug, Clone, PartialEq, Eq, Hash)]
+#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
 pub struct DebugSymbol {
     pub start: TokenIndex,
     pub end: TokenIndex,
@@ -97,20 +97,9 @@ impl ResolvedSymbol {
 
 }
 
-impl fmt::Display for ResolvedSymbol {
-    fn fmt(&self, fmt: &mut fmt::Formatter<'_>) -> fmt::Result {
-        for line in self.iter_lines() {
-            fmt.write_str(line.trim_end())?;
-            fmt.write_str("\n")?;
-        }
-        Ok(())
-    }
-}
-
-
-// Resolved Symbol Formatting
 
 // Symbol Resolution
+// TODO move to sub-module
 
 pub type ResolvedSymbolTable<'s> = HashMap<&'s DebugSymbol, Result<ResolvedSymbol, SymbolResolutionError>>;
 
@@ -121,21 +110,63 @@ pub trait DebugSymbolResolver {
 impl DebugSymbolResolver for ModuleSource {
     fn resolve_symbols<'s, S>(&self, symbols: S) -> io::Result<ResolvedSymbolTable<'s>> where S: Iterator<Item=&'s DebugSymbol> {
         match self.source_text()? {
-            SourceText::String { text, .. } => {
-                Ok(resolve_debug_symbols(text.chars().map(Ok), symbols))
-            },
-            SourceText::File { text, .. } => {
-                Ok(resolve_debug_symbols(text, symbols))
-            }
+            SourceText::String(text) => Ok(resolve_debug_symbols(text.chars().map(Ok), symbols)),
+            SourceText::File(text) => Ok(resolve_debug_symbols(text, symbols))
         }
     }
 }
 
+pub struct BufferedResolver {
+    buffer: String,
+}
+
+impl BufferedResolver {
+    pub fn new(string: impl ToString) -> Self {
+        Self { buffer: string.to_string() }
+    }
+}
+
+impl DebugSymbolResolver for BufferedResolver {
+    fn resolve_symbols<'s, S>(&self, symbols: S) -> io::Result<ResolvedSymbolTable<'s>> where S: Iterator<Item=&'s DebugSymbol> {
+        Ok(resolve_debug_symbols(self.buffer.chars().map(Ok), symbols))
+    }
+}
+
+// use std::fs;
+// use std::path::PathBuf;
+// use crate::utils::ReadChars;
+
+// /// Resolves 
+// pub struct FileSymbolResolver {
+//     path: PathBuf,
+// }
+
+// impl FileSymbolResolver {
+//     pub fn new(path: impl Into<PathBuf>) -> Self { 
+//         Self { path: path.into() } 
+//     }
+// }
+
+// impl DebugSymbolResolver for FileSymbolResolver {
+//     fn resolve_symbols<'s, S>(&self, symbols: S) -> io::Result<ResolvedSymbolTable<'s>> where S: Iterator<Item=&'s DebugSymbol> {
+//         let file = fs::File::open(self.path)?;
+//         let reader = io::BufReader::new(file);
+//         let reader = ReadChars::new(reader);
+//         Ok(resolve_debug_symbols(reader, symbols))
+//     }
+// }
+
+// resolution procedure
+
 fn resolve_debug_symbols<'s>(source: impl Iterator<Item=io::Result<char>>, symbols: impl Iterator<Item=&'s DebugSymbol>) -> ResolvedSymbolTable<'s> {
+    // deduplicate symbols
+    let mut dedup_symbols = symbols.collect::<Vec<&DebugSymbol>>();
+    dedup_symbols.dedup();
+    
     // put all the symbols into a priority queue based on first occurrence in the source text
     let mut next_symbols = BinaryHeap::new();
-    next_symbols.extend(symbols.map(|sym| IndexSort(sym, SortIndex::Start)).map(cmp::Reverse));
-    let symbol_count = next_symbols.len();
+    next_symbols.extend(dedup_symbols.into_iter().map(|sym| IndexSort(sym, SortIndex::Start)).map(cmp::Reverse));
+    // let symbol_count = next_symbols.len();
     
     let mut open_symbols = BinaryHeap::new();
     let mut active_symbols = HashMap::<&DebugSymbol, (Vec<Rc<String>>, usize, usize)>::new(); // values are (buffer, line number, start index)
@@ -152,11 +183,14 @@ fn resolve_debug_symbols<'s>(source: impl Iterator<Item=io::Result<char>>, symbo
         let c = match char_result {
             Ok(c) => c,
             
-            Err(..) => {
+            Err(ioerror) => {
+                let ioerror = Rc::new(ioerror);
+                
                 // drop all open symbols and close all closing symbols
                 active_symbols.clear();
                 for cmp::Reverse(IndexSort(symbol,..)) in open_symbols.drain() {
-                    resolved_symbols.insert(symbol, Err(ErrorKind::IOError.into()));
+                    let error = SymbolResolutionError::caused_by(ErrorKind::IOError, *symbol, ioerror.clone());
+                    resolved_symbols.insert(symbol, Err(error));
                 }
                 
                 // we've already have the required text for all closing symbols, so we don't need to drop them
@@ -213,12 +247,12 @@ fn resolve_debug_symbols<'s>(source: impl Iterator<Item=io::Result<char>>, symbo
             
             // open symbols
             for (ref mut lines, ..) in active_symbols.values_mut() {
-                lines.push(line_rc.clone());
+                lines.push(Rc::clone(&line_rc));
             }
             
             // close all closing symbols
             for (symbol, (mut lines, lineno, start_index, end_index)) in closing_symbols.drain() {
-                lines.push(line_rc.clone());
+                lines.push(Rc::clone(&line_rc));
                 
                 let resolved = ResolvedSymbol::new(lines, lineno, start_index, end_index);
                 resolved_symbols.insert(symbol, Ok(resolved));
@@ -242,7 +276,8 @@ fn resolve_debug_symbols<'s>(source: impl Iterator<Item=io::Result<char>>, symbo
     
     // process any active or closing symbols left after we hit EOF
     for cmp::Reverse(IndexSort(symbol,..)) in open_symbols.drain() {
-        resolved_symbols.insert(symbol, Err(ErrorKind::EOFReached.into()));
+        let error = SymbolResolutionError::new(ErrorKind::EOFReached, *symbol);
+        resolved_symbols.insert(symbol, Err(error));
     }
     
     let line_rc = Rc::new(current_line.clone());
@@ -253,7 +288,7 @@ fn resolve_debug_symbols<'s>(source: impl Iterator<Item=io::Result<char>>, symbo
         resolved_symbols.insert(symbol, Ok(resolved));
     }
     
-    debug_assert!(symbol_count == resolved_symbols.len());
+    // debug_assert!(symbol_count == resolved_symbols.len());
     
     resolved_symbols
 }
@@ -263,6 +298,7 @@ fn resolve_debug_symbols<'s>(source: impl Iterator<Item=io::Result<char>>, symbo
 enum SortIndex { Start, End }
 
 // comparison based on start or end index
+
 #[derive(Debug)]
 struct IndexSort<'s>(&'s DebugSymbol, SortIndex);
 
@@ -298,10 +334,8 @@ impl Ord for IndexSort<'_> {
 
 use std::error::Error;
 
-pub type ErrorKind = SymbolResolutionErrorKind;
-
 #[derive(Debug)]
-pub enum SymbolResolutionErrorKind {
+pub enum ErrorKind {
     IOError,    // hit an io::Error while grabbing the symbol text
     EOFReached, // hit EOF before reaching the indicated end of symbol
 }
@@ -309,31 +343,26 @@ pub enum SymbolResolutionErrorKind {
 #[derive(Debug)]
 pub struct SymbolResolutionError {
     kind: ErrorKind,
-    cause: Option<Box<dyn std::error::Error>>,
+    symbol: DebugSymbol,
+    cause: Option<Rc<dyn std::error::Error>>,
 }
 
 impl SymbolResolutionError {
-    pub fn new(kind: ErrorKind) -> Self {
-        SymbolResolutionError {
-            kind, cause: None,
+    pub fn new(kind: ErrorKind, symbol: DebugSymbol) -> Self {
+        Self {
+            kind, symbol, cause: None,
         }
     }
     
-    pub fn caused_by(kind: ErrorKind, error: impl std::error::Error + 'static) -> Self {
-        SymbolResolutionError {
-            kind, cause: Some(Box::new(error)),
+    pub fn caused_by(kind: ErrorKind, symbol: DebugSymbol, error: Rc<dyn std::error::Error>) -> Self {
+        Self {
+            kind, symbol, cause: Some(error),
         }
     }
     
     pub fn kind(&self) -> &ErrorKind { &self.kind }
 }
 
-impl From<ErrorKind> for SymbolResolutionError {
-    fn from(kind: ErrorKind) -> Self {
-        Self::new(kind)
-    }
-}
-
 impl Error for SymbolResolutionError {
     fn source(&self) -> Option<&(dyn std::error::Error + 'static)> {
         self.cause.as_ref().map(|o| o.as_ref())
@@ -343,11 +372,12 @@ impl Error for SymbolResolutionError {
 impl fmt::Display for SymbolResolutionError {
     fn fmt(&self, fmt: &mut fmt::Formatter<'_>) -> fmt::Result {
         
-        let message = match self.kind() {
-            SymbolResolutionErrorKind::IOError => "I/O error while extracting symbol",
-            SymbolResolutionErrorKind::EOFReached => "EOF reached while extracting symbol",
+        let inner_message = match self.kind() {
+            ErrorKind::IOError => "I/O error while extracting symbol",
+            ErrorKind::EOFReached => "EOF reached while extracting symbol",
         };
         
-        utils::format_error(fmt, "could not resolve symbol", Some(message), self.source())
+        let message = format!("could not resolve symbol ${}:{}", self.symbol.start, self.symbol.end);
+        utils::format_error(fmt, message.as_str(), Some(inner_message), self.source())
     }
 }
\ No newline at end of file
