@@ -59,10 +59,11 @@ pub fn create_default_lexer_rules() -> LexerBuilder {
     .add_rule(SingleCharRule::new(Token::OpMul,           '*'))
     .add_rule(SingleCharRule::new(Token::OpDiv,           '/'))
     .add_rule(SingleCharRule::new(Token::OpMod,           '%'))
+    .add_rule(SingleCharRule::new(Token::OpExp,           '^'))
     .add_rule(SingleCharRule::new(Token::OpInv,           '~'))
     .add_rule(SingleCharRule::new(Token::OpAnd,           '&'))
     .add_rule(SingleCharRule::new(Token::OpOr,            '|'))
-    .add_rule(SingleCharRule::new(Token::OpXor,           '^'))
+    .add_rule(SingleCharRule::new(Token::OpXor,           '~'))
     
     .add_rule(SingleCharRule::new(Token::OpLT,            '<'))
     .add_rule(SingleCharRule::new(Token::OpGT,            '>'))
@@ -95,6 +96,7 @@ pub fn create_default_lexer_rules() -> LexerBuilder {
     .add_rule(KeywordRule::new(Token::Nil,                "nil"))
     .add_rule(KeywordRule::new(Token::Let,                "let"))
     .add_rule(KeywordRule::new(Token::Var,                "var"))
+    .add_rule(KeywordRule::new(Token::NonLocal,           "nonlocal"))
     .add_rule(KeywordRule::new(Token::Del,                "del"))
     .add_rule(KeywordRule::new(Token::Begin,              "begin"))
     .add_rule(KeywordRule::new(Token::If,                 "if"))
@@ -102,6 +104,7 @@ pub fn create_default_lexer_rules() -> LexerBuilder {
     .add_rule(KeywordRule::new(Token::Elif,               "elif"))
     .add_rule(KeywordRule::new(Token::Else,               "else"))
     .add_rule(KeywordRule::new(Token::For,                "for"))
+    .add_rule(KeywordRule::new(Token::In,                 "in"))
     .add_rule(KeywordRule::new(Token::While,              "while"))
     .add_rule(KeywordRule::new(Token::Do,                 "do"))
     .add_rule(KeywordRule::new(Token::Continue,           "continue"))
@@ -109,9 +112,10 @@ pub fn create_default_lexer_rules() -> LexerBuilder {
     .add_rule(KeywordRule::new(Token::Return,             "return"))
     .add_rule(KeywordRule::new(Token::Fun,                "fun"))
     .add_rule(KeywordRule::new(Token::Class,              "class"))
-    .add_rule(KeywordRule::new(Token::Self_,              "self"))
-    .add_rule(KeywordRule::new(Token::Super,              "super"))
+    // .add_rule(KeywordRule::new(Token::Self_,              "self"))
+    // .add_rule(KeywordRule::new(Token::Super,              "super"))
     .add_rule(KeywordRule::new(Token::Echo,               "echo"))
+    .add_rule(KeywordRule::new(Token::Assert,             "assert"))
     .add_rule(KeywordRule::new(Token::End,                "end"))
     
     // Identifiers and literals
@@ -122,5 +126,4 @@ pub fn create_default_lexer_rules() -> LexerBuilder {
     .add_rule(StringLiteralRule::new(ESCAPE_SEQUENCES.iter().map(|esc| esc.as_ref())))
     .add_rule(LabelRule::new("::"))
     
-
 }
