@@ -4,7 +4,7 @@ use crate::parser::primary::{Atom, Primary};
 use crate::parser::assign::{Assignment, Declaration};
 use crate::parser::fundefs::FunctionDef;
 use crate::parser::structs::{ObjectConstructor};
-use crate::parser::stmt::{StmtMeta, Stmt, Label};
+use crate::parser::stmt::{StmtMeta, Stmt, Label, StmtList};
 
 // TODO replace Vecs with boxed slices
 #[derive(Debug, Clone)]
@@ -26,9 +26,9 @@ pub enum Expr {
     
     // ObjectCtor(Box<ObjectConstructor>),
     
-    IfExpr(IfExpr),
+    IfExpr(Conditional),
     
-    Block(Option<Label>, Box<[StmtMeta]>),
+    Block(Option<Label>, StmtList),
     
     FunctionDef(FunctionDef),
     
@@ -36,15 +36,33 @@ pub enum Expr {
 }
 
 #[derive(Debug, Clone)]
-pub struct IfExpr {
-    pub branches: Box<[CondBranch]>,
-    pub else_branch: Option<Box<[StmtMeta]>>,
+pub struct Conditional {
+    branches: Box<[CondBranch]>,
+    else_branch: Option<StmtList>,
+}
+
+impl Conditional {
+    pub fn new(branches: Vec<CondBranch>, else_branch: Option<StmtList>) -> Self {
+        Self {
+            branches: branches.into_boxed_slice(),
+            else_branch,
+        }
+    }
+    
+    pub fn branches(&self) -> &[CondBranch] { &self.branches }
+    pub fn else_branch(&self) -> Option<&StmtList> { self.else_branch.as_ref() }
 }
 
 #[derive(Debug, Clone)]
-pub struct CondBranch {
-    pub cond: Expr,
-    pub suite: Box<[StmtMeta]>,
+pub struct CondBranch(Expr, StmtList);
+
+impl CondBranch {
+    pub fn new(cond: Expr, stmt_list: StmtList) -> Self {
+        Self(cond, stmt_list)
+    }
+    
+    pub fn cond_expr(&self) -> &Expr { &self.0 }
+    pub fn suite(&self) -> &StmtList { &self.1 }
 }
 
 
@@ -83,12 +101,23 @@ impl From<Expr> for Stmt {
 }
 
 impl TryFrom<Stmt> for Expr {
-    type Error = ();
+    type Error = Stmt;
     
     #[inline]
-    fn try_from(stmt: Stmt) -> Result<Self, Self::Error> {
+    fn try_from(stmt: Stmt) -> Result<Self, Stmt> {
         if let Stmt::Expression(expr) = stmt { Ok(expr) }
-        else { Err(()) }
+        else { Err(stmt) }
+    }
+}
+
+impl TryFrom<StmtMeta> for Expr {
+    type Error = StmtMeta;
+    
+    #[inline]
+    fn try_from(stmt: StmtMeta) -> Result<Self, StmtMeta> {
+        let (stmt, symbol) = stmt.take();
+        if let Stmt::Expression(expr) = stmt { Ok(expr) }
+        else { Err(StmtMeta::new(stmt, symbol)) }
     }
 }
 
@@ -102,11 +131,14 @@ impl From<ExprMeta> for StmtMeta {
 }
 
 impl TryFrom<StmtMeta> for ExprMeta {
-    type Error = ();
+    type Error = StmtMeta;
     
     #[inline]
-    fn try_from(stmt: StmtMeta) -> Result<Self, Self::Error> {
+    fn try_from(stmt: StmtMeta) -> Result<Self, StmtMeta> {
         let (stmt, symbol) = stmt.take();
-        Ok(ExprMeta::new(stmt.try_into()?, symbol))
+        match Expr::try_from(stmt) {
+            Ok(expr) => Ok(ExprMeta::new(expr, symbol)),
+            Err(stmt) => Err(StmtMeta::new(stmt, symbol)),
+        }
     }
 }
\ No newline at end of file
