@@ -1,6 +1,6 @@
 use crate::runtime::strings::InternSymbol;
 use crate::debug::DebugSymbol;
-use crate::parser::expr::Expr;
+use crate::parser::expr::{Expr, ExprMeta};
 
 
 #[derive(Debug, Clone, Copy, PartialEq, Eq)]
@@ -11,7 +11,6 @@ impl Label {
     pub fn name(&self) -> &InternSymbol { &self.0 }
 }
 
-
 #[derive(Debug, Clone)]
 pub enum Stmt {
     
@@ -21,15 +20,42 @@ pub enum Stmt {
     // DoWhileLoop
     // ForLoop
     
-    Continue(Option<Label>),
-    Break(Option<Label>, Option<Expr>),
-    Return(Option<Expr>),
-    
     Echo(Expr),
     Assert(Expr),
 }
 
-impl Stmt { }
+
+#[derive(Debug, Clone)]
+pub enum ControlFlow {
+    Continue(Option<Label>),
+    Break(Option<Label>, Option<Box<Expr>>),
+    Return(Option<Box<Expr>>),
+    Expression(Box<ExprMeta>),
+}
+
+
+#[derive(Debug, Clone)]
+pub struct StmtList {
+    suite: Box<[StmtMeta]>,
+    control: Option<ControlFlow>,
+}
+
+impl StmtList {
+    pub fn new(suite: Vec<StmtMeta>, control: Option<ControlFlow>) -> Self {
+        Self {
+            suite: suite.into_boxed_slice(),
+            control,
+        }
+    }
+    
+    pub fn suite(&self) -> &[StmtMeta] { &self.suite }
+    pub fn control(&self) -> Option<&ControlFlow> { self.control.as_ref() }
+    
+    pub fn take(self) -> (Vec<StmtMeta>, Option<ControlFlow>) {
+        (self.suite.into_vec(), self.control)
+    }
+}
+
 
 // Stmt + DebugSymbol
 #[derive(Debug, Clone)]
