@@ -1,6 +1,4 @@
 // Data structures
 
-use crate::parser::expr::ExprMeta;
-
-#[derive(Debug, Clone)]
+#[derive(Clone)]
 pub struct ObjectConstructor { }  // TODO
